# Code dedicated to the postprocessing new heads check with obsolescence
#
# Copyright 2017 Pierre-Yves David <pierre-yves.david@ens-lyon.org>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from mercurial import (
    discovery,
    error,
    extensions,
    node as nodemod,
    obsolete,
    phases,
    util,
)

from mercurial.i18n import _

from . import exthelper

nullid = nodemod.nullid
short = nodemod.short
_headssummary = discovery._headssummary
_oldheadssummary = discovery._oldheadssummary
_nowarnheads = discovery._nowarnheads

eh = exthelper.exthelper()

@eh.uisetup
def setupcheckheadswrapper(ui):
    if util.safehasattr(discovery, '_postprocessobsolete'):
        extensions.wrapfunction(discovery, '_postprocessobsolete',
                                checkheadslightoverlay)
    else:
        extensions.wrapfunction(discovery, 'checkheads',
                                checkheadsfulloverlay)

# have dedicated wrapper to keep the rest as close as core as possible
def checkheadsfulloverlay(orig, pushop):
    if pushop.repo.obsstore:
        return corecheckheads(pushop)
    else:
        return orig(pushop)

def checkheadslightoverlay(orig, *args, **kwargs):
    return _postprocessobsolete(*args, **kwargs)

# copied from mercurial.discovery.checkheads as in a5bad127128d (4.1)
#
# The only differences are:
# * the _postprocessobsolete section have been extracted,
# * minor test adjustment to please flake8
def corecheckheads(pushop):
    """Check that a push won't add any outgoing head

    raise Abort error and display ui message as needed.
    """

    repo = pushop.repo.unfiltered()
    remote = pushop.remote
    outgoing = pushop.outgoing
    remoteheads = pushop.remoteheads
    newbranch = pushop.newbranch
    inc = bool(pushop.incoming)

    # Check for each named branch if we're creating new remote heads.
    # To be a remote head after push, node must be either:
    # - unknown locally
    # - a local outgoing head descended from update
    # - a remote head that's known locally and not
    #   ancestral to an outgoing head
    if remoteheads == [nullid]:
        # remote is empty, nothing to check.
        return

    if remote.capable('branchmap'):
        headssum = _headssummary(repo, remote, outgoing)
    else:
        headssum = _oldheadssummary(repo, remoteheads, outgoing, inc)
    newbranches = [branch for branch, heads in headssum.iteritems()
                   if heads[0] is None]
    # 1. Check for new branches on the remote.
    if newbranches and not newbranch:  # new branch requires --new-branch
        branchnames = ', '.join(sorted(newbranches))
        raise error.Abort(_("push creates new remote branches: %s!")
                          % branchnames,
                          hint=_("use 'hg push --new-branch' to create"
                                 " new remote branches"))

    # 2. Find heads that we need not warn about
    nowarnheads = _nowarnheads(pushop)

    # 3. Check for new heads.
    # If there are more heads after the push than before, a suitable
    # error message, depending on unsynced status, is displayed.
    errormsg = None
    # If there is no obsstore, allfuturecommon won't be used, so no
    # need to compute it.
    if repo.obsstore:
        allmissing = set(outgoing.missing)
        cctx = repo.set('%ld', outgoing.common)
        allfuturecommon = set(c.node() for c in cctx)
        allfuturecommon.update(allmissing)
    for branch, heads in sorted(headssum.iteritems()):
        remoteheads, newheads, unsyncedheads = heads
        candidate_newhs = set(newheads)
        # add unsynced data
        if remoteheads is None:
            oldhs = set()
        else:
            oldhs = set(remoteheads)
        oldhs.update(unsyncedheads)
        candidate_newhs.update(unsyncedheads)
        dhs = None # delta heads, the new heads on branch
        if not repo.obsstore:
            discardedheads = set()
            newhs = candidate_newhs
        else:
            newhs, discardedheads = _postprocessobsolete(pushop,
                                                         allfuturecommon,
                                                         candidate_newhs)
        unsynced = sorted(h for h in unsyncedheads if h not in discardedheads)
        if unsynced:
            if None in unsynced:
                # old remote, no heads data
                heads = None
            elif len(unsynced) <= 4 or repo.ui.verbose:
                heads = ' '.join(short(h) for h in unsynced)
            else:
                heads = (' '.join(short(h) for h in unsynced[:4]) +
                         ' ' + _("and %s others") % (len(unsynced) - 4))
            if heads is None:
                repo.ui.status(_("remote has heads that are "
                                 "not known locally\n"))
            elif branch is None:
                repo.ui.status(_("remote has heads that are "
                                 "not known locally: %s\n") % heads)
            else:
                repo.ui.status(_("remote has heads on branch '%s' that are "
                                 "not known locally: %s\n") % (branch, heads))
        if remoteheads is None:
            if len(newhs) > 1:
                dhs = list(newhs)
                if errormsg is None:
                    errormsg = (_("push creates new branch '%s' "
                                  "with multiple heads") % (branch))
                    hint = _("merge or"
                             " see 'hg help push' for details about"
                             " pushing new heads")
        elif len(newhs) > len(oldhs):
            # remove bookmarked or existing remote heads from the new heads list
            dhs = sorted(newhs - nowarnheads - oldhs)
        if dhs:
            if errormsg is None:
                if branch not in ('default', None):
                    errormsg = _("push creates new remote head %s "
                                 "on branch '%s'!") % (short(dhs[0]), branch)
                elif repo[dhs[0]].bookmarks():
                    errormsg = (_("push creates new remote head %s "
                                  "with bookmark '%s'!")
                                % (short(dhs[0]), repo[dhs[0]].bookmarks()[0]))
                else:
                    errormsg = _("push creates new remote head %s!"
                                 ) % short(dhs[0])
                if unsyncedheads:
                    hint = _("pull and merge or"
                             " see 'hg help push' for details about"
                             " pushing new heads")
                else:
                    hint = _("merge or"
                             " see 'hg help push' for details about"
                             " pushing new heads")
            if branch is None:
                repo.ui.note(_("new remote heads:\n"))
            else:
                repo.ui.note(_("new remote heads on branch '%s':\n") % branch)
            for h in dhs:
                repo.ui.note((" %s\n") % short(h))
    if errormsg:
        raise error.Abort(errormsg, hint=hint)

def _postprocessobsolete(pushop, futurecommon, candidate):
    """post process the list of new heads with obsolescence information

    Exist as a subfunction to contains the complexity and allow extensions to
    experiment with smarter logic.
    Returns (newheads, discarded_heads) tuple
    """
    # remove future heads which are actually obsoleted by another
    # pushed element:
    #
    # XXX as above, There are several cases this code does not handle
    # XXX properly. some below:
    #
    # * We "silently" skip processing on all changeset unknown locally
    #
    # * if <nh> is public on the remote, it won't be affected by obsolete
    #     marker and a new is created
    #
    # * if a successors is pruned, but the prune marker will not be exchanged,
    #   we currently badely detect the situation.
    #
    # These two cases will be easy to handle for known changeset but
    # much more tricky for unsynced changes.
    repo = pushop.repo
    unfi = repo.unfiltered()
    cl = unfi.changelog
    newhs = set()
    discarded = set()
    # I leave the print in the code because they are so handy at debugging
    # and I keep getting back to this piece of code.
    #
    # from mercurial.node import short
    localcandidate = set()
    unknownheads = set()
    for h in candidate:
        if h in unfi:
            localcandidate.add(h)
        else:
            unknownheads.add(h)
    if len(localcandidate) == 1:
        return unknownheads | set(candidate), set()
    while localcandidate:
        nh = localcandidate.pop()
        # print 'newhead:', short(nh)
        currentbranch = unfi.revs('only(%n, (%ln+%ln))',
                                  nh, localcandidate, newhs)
        if nh in futurecommon:
            # The heads will still be alive in the future set
            # print ' >> head is common'
            newhs.add(nh)
            continue
        stack = [nh]
        seen = set(stack)
        while stack:
            current = stack.pop()
            # print ' current:', short(current)
            if current in futurecommon:
                if cl.rev(current) not in currentbranch:
                    # we reach the bottom of the branch, do not go further
                    # print '  > futurecommon'
                    continue
            # XXX we do not know the phase on the remote side
            if unfi[current].phase() <= phases.public:
                # public changeset cannot be rewritten. This head will remain.
                newhs.add(nh)
                break
            # We check is the changeset will disappear on push. It happens when
            #
            # * the changeset is pruned,
            # * the changeset is superceed by a one in future common set.
            #
            # XXX note: maybe we can just check is the changeset is affected by a
            # markers relevant to the 'push-set' but that. It might be simpler

            # XXX we need some "stop at" argument to 'successorssets' to
            # XXX prevent iterating too far. But pushing obsolete changeset
            # XXX will requires --force anyway.
            sucsets = obsolete.successorssets(unfi, current)
            # print '  =', short(current), [tuple(short(s) for s in ss) for ss in sucsets]
            if sucsets == []: # pruned
                # XXX possibly, the prune markers is further away and rooted on
                # some changeset we do not push

                # print '  - pruned'
                # XXX if current is unknown, we should request a pull
                if current in unfi:
                    for p in unfi[current].parents():
                        pnode = p.node()
                        if pnode not in seen:
                            seen.add(pnode)
                            stack.append(pnode)
                continue
            for ss in sucsets:
                if ss == (current,): # changeset is alive
                    # print '  < alive'
                    newhs.add(nh)
                    break
                for suc in ss:
                    # XXX we should do something special for split
                    # split -> 1 < len(ss)
                    if suc in futurecommon:
                        # print '  - superceeded (%s)' % (current in repo)
                        if current in unfi:
                            for p in unfi[current].parents():
                                pnode = p.node()
                                if pnode not in seen:
                                    seen.add(pnode)
                                    stack.append(pnode)
                        # XXX if current is unknown, we should request a pull
                        break
                else:
                    continue
                break # propagate the break
            else:
                for ss in sucsets:
                    for suc in ss:
                        if suc != nh and suc in futurecommon:
                            discarded.add(nh)
                            break
                    else:
                        continue
                    break # propagate the break
                else:
                    newhs.add(nh)
    newhs |= unknownheads
    return newhs, discarded
